-- Misc shearing- and shadow-related demos

module Shadow where

import Monad
import Pan hiding (softenUI)

-- ToDo: reconcile Neighbor.soften with Shadow.soften

---- Shears and shadowed images

-- Horizontal shearing, given base (invariant) height
shearUI :: FloatE -> UI (Filter a)
shearUI base = do ang <- angleUI "shear angle" (-225,225) 30
                  str <- slider  "shear scale" (0.01,2) 0.8
                  let (scx,scy) = fromPolar (str,ang+pi/2)
                  return $ about (0,base) (scale (1,scy) . shear scx)

shearTest :: FloatE -> Image a -> UI (Image a)
shearTest base im = liftM ($ im) (shearUI base)

shearedDiskUI = liftM ($ diskIm rad red) (shearUI (-rad))
                where rad = 30

-- Color in shadow image is shadow color, faded by image's alpha
shadow :: ColorE -> FilterC
shadow shadowC im = fade (alphaI im) (const shadowC)

-- Produce a shadow
shadowUI :: ColorE -> FloatE -> ImageC -> UI ImageC
shadowUI shadowC base im =
  do shearF  <- shearUI base
     opacity <- slider "shadow opacity" (0,1) 0.5
     return $ fade (const opacity) (shearF (shadow shadowC im))

-- Combine image and shadow.  Sometimes more convenient.
addShadowUI :: ColorE -> FloatE -> ImageC -> UI ImageC
addShadowUI shadowC base im = liftM (im `over`) (shadowUI shadowC base im)

shadowedDiskUI = addShadowUI black (-25) (diskIm 25 red)

shadowedBoxUI = do (w,h) <- sizeUI
                   addShadowUI black (-h/2) (crop (box w h) blueI)


-- Some tools for soft shadows

softenUI :: Array2 ColorE -> UI ImageC
softenUI bmp = do softness <- slider "softness" (0,10) 1
                  let ker = uscale softness (gauss (1/2) . distOsq)
                  return $ convolveDC (roundE softness) ker bmp

{-
-- Worth factoring out?
-- Soft shadow of given discrete image
softShadowUI :: ColorE -> FloatE -> Array2 ColorE -> UI ImageC
softShadowUI shadowC base bmp =
  softIm <- softenUI bmp
  shadowUI shadowC base softIm
-}

addSoftShadowUI :: ColorE -> FloatE -> Array2 ColorE -> UI ImageC
addSoftShadowUI shadowC base bmp =
  do softIm <- softenUI bmp
     shadow <- shadowUI shadowC base softIm
     return $ reconstruct bmp `over` shadow



-- Two methods for adding alpha.  One maps lightness to transparency, and
-- is good for anti-aliased black text on a white background.  The other
-- uses chroma-keying with white.  Could be generalized, but better would
-- be to use bitmaps with alpha, e.g., PNG.

-- Continuous versions
lightBg :: ImageC -> ImageC
lightBg = transpLightI

whiteBg :: FractionE -> ImageC -> ImageC
whiteBg eps = chromaKey eps white

-- Discrete versions
lightBgD :: Array2 ColorE -> Array2 ColorE
lightBgD = arr2Map $ \ col -> fadeC (darkness col) col

whiteBgD :: FractionE -> Array2 ColorE -> Array2 ColorE
whiteBgD eps = arr2Map $ \ col -> ifE (withinC eps white col) invisible col


shadowedBmpUI bmp addAlpha =
  do let base = - i2f (arr2Height bmp) / 2
     withShadow <- addSoftShadowUI black base (addAlpha bmp)
     return $ withShadow `over` ybRings

shadowedBmpUI' bmp addAlpha =
  do let base = - i2f (arr2Height bmp) / 2
     addSoftShadowUI black base (addAlpha bmp)

shadowedImportUI addAlpha =
  do bmp <- imageSourceD "Image"
     shadowedBmpUI bmp addAlpha

shadowedChromaUI, shadowedLightnessUI :: UI ImageC

shadowedChromaUI =
  do eps <- slider "chroma tolerance" (0,1.5) 0.01
     shadowedImportUI (whiteBgD eps)

shadowedLightnessUI = shadowedImportUI lightBgD

{-
shadowedChromaUI =
  do (w,h,im)   <- imageSourceSize "Image"
     eps        <- slider "chroma tolerance" (0,1.5) 0.01
     withShadow <- addSoftShadowUI black (-h/2) (whiteBg eps im)
     return $ withShadow `over` ybRings

shadowedLightnessUI =
  do (w,h,im)   <- imageSourceSize "Image"
     eps        <- slider "chroma tolerance" (0,1.5) 0.01
     withShadow <- addSoftShadowUI black (-h/2) (lightBgim)
     return $ withShadow `over` ybRings

-- This one allows the user to select transparency style.
shadowedSelectUI =
  do (w,h,im)   <- browseImageWithAlphaSize "Image" "PhotoDraw.jpg"
     withShadow <- addSoftShadowUI blue (-h/2) im
     return $ withShadow `over` ybRings
-}


alphaTestUI = do im <- browseImageWithAlpha "Image" "Becky.bmp"
                 return $ im `over` cb0



----------------------------------------

makeAll = makeAllFor allBackends

makeAllFor bends = do
  compileImageFor bends "sheared disk"       shearedDiskUI
  compileImageFor bends "shadowed disk"      shadowedDiskUI
  compileImageFor bends "shadowed box"       shadowedBoxUI
  compileImageFor bends "shadowed lightness" shadowedLightnessUI
  compileImageFor bends "shadowed chroma"    shadowedChromaUI
