
Copyright (c) 1998,  Daan Leijen, leijen@@fwi.uva.nl

This module is part of HaskellDirect (H/Direct).

\begin{code}
module Pointer  
	( 
	  Ptr
        , nullPtr
	, castPtr
	
	, allocMemory
	, stackFrame
	, ptrInc

	, writeSeqAt
	, writeSeqAtDec

        , makePtr
        , makeCPtr
	, makeNoFreePtr

        , freeMemory
	, freeBSTR
	, freeWith
	, freeWithC
	
	, primNoFree
	
	, finalNoFree
	, finalFreeMemory
	
	, makeFO

       ) where

import IOExts   ( unsafePerformIO )
import Int
{- BEGIN_GHC_ONLY 
import GlaExts
   END_GHC_ONLY -}
import Foreign
import Addr
import PointerPrim  -- look here for Pointer type defn.
import Word     ( Word32 )
import Weak
import Monad
\end{code}

The @Pointer@ data type is a typed pointer to something
outside of Haskell. The pointer is finalised, that is,
when the garbage collector has determined a @Pointer@
to be unreachable, it calls the supplied finaliser
to notify the outside world that we're through with
using the resource it points to.

Pointers are created with @makePointer@ which takes a (Haskell)
finalizer routine and an address. Since foreign objects are not
yet completely supported, it is not possible to change the finalizer
routine.  The finalizer routine is also not completely supported and
is an addr to a C finalizer routine at the moment. 

Special creation routines are:  

  - @makeInterfacePointer@
             calls @release@ on a COM object,
  - @makeCPointer@ 
             calls @free@ (or @CoTaskMemFree@) on a C pointer,
  - @makeNoFreePointer@ 
             doesn't call a finalizer.

Pointers can be freed manually if it is statically known that
some memory can be freed. Use @freePointer@ with a finalizer routine
or call the specialized @freeMemory@. (The pointer will NOT
be freed again on GC).

\begin{code}

type Ptr a	  = Addr
type Finalizer a  = Ptr a -> IO ()

makePtr       :: Finalizer a -> Addr -> IO (Ptr a)
makePtr free addr  = do
   addFinalizer addr (free addr)
   return addr

makeFO :: Addr -> Addr -> IO ForeignObj
makeFO obj finaliser = makeForeignObj obj finaliser

makeNoFreePtr    :: Addr -> IO (Ptr a)
makeCPtr         :: Addr -> IO (Ptr a)

makeNoFreePtr       = makePtr primNoFree
makeCPtr            = makePtr primFreeMemory

castPtr :: Ptr a -> Ptr b
castPtr p = p

\end{code}

Helpers.

\begin{code}
ptrInc i p        = p `plusAddr` (toInt i)
ptrDec d p        = p `plusAddr` (-(toInt  d))

writeSeqAt :: Word32 -> [Ptr a -> IO ()] -> Ptr a -> IO ()
writeSeqAt size ws ptr = go ptr ws
 where
   go ptr []     = return ()
   go ptr (x:xs) = do
      x ptr
      let ptr_next = ptrInc size ptr
      go ptr_next xs

writeSeqAtDec :: Word32 -> [Ptr a -> IO ()] -> Ptr a -> IO ()
writeSeqAtDec size ws ptr = go init_ptr ws
  where
   len           = fromInt (length ws - 1)
   init_ptr      = ptrInc  (size * len) ptr

   go ptr []     = return ()
   go ptr (x:xs) = do
      x ptr
      let ptr_next = ptrDec size ptr
      go ptr_next xs

\end{code}

Use @stackFrame@ when you know the allocated chunk have
a fixed extent.

\begin{code}
allocStack    = allocFrame

stackFrame :: Word32 -> (Ptr a -> IO b) -> IO b
stackFrame size f
      = do p <- allocFrame size
           f p `always` primFreeMemory p

\end{code}

Special free routines for pointers. Use them to manually free pointers.

\begin{code}
freeMemory            = freeWithC primFreeMemory
freeBSTR              = freeWithC primFreeBSTR

freeWithC :: Finalizer a -> Ptr a -> IO ()
freeWithC final p = final p

freeWith :: (Ptr a -> IO ()) -> Ptr a -> IO ()
freeWith free p = free p 
\end{code}

Helper functions that doesn't really have a good home to go to:

\begin{code}
always :: IO a -> IO () -> IO a
always io action
      = do x <- io `catch` (\ err -> do { action; ioError err })
           action
           return x

\end{code}

Primitives/helpers:

\begin{code}
nullPtr :: Ptr a
nullPtr = nullAddr

allocMemory :: Word32 -> IO (Ptr a)
allocMemory sz = do
  a <- primAllocMemory sz
  if a == nullAddr then
     ioError (userError "allocMemory: not enough memory")
   else
     return a

allocFrame :: Word32 -> IO (Ptr a)
allocFrame sz = do
  a <- primAllocFrame sz
  if a == nullAddr then
     ioError (userError "allocFrame: not enough memory")
   else
     return a

\end{code}
