-- Statically typed expression wrappers.  Built on DExp.hs.

module Exp where

import Int
import Bits hiding (rotate)
import IOExts (trace)

import qualified F32vec4 as V4
import PP
import Types
import qualified DExp as D
import qualified SmartDExp as SD
import qualified PPDExp ( ppExp,showsPrecDExp )
import TCheck (TCheck, tck, tcheck)
import IOExts ( unsafePerformIO )

infixl 8 <<<, >>>  -- left and right shift
infix  4  ==*, /=*, <*, <=*, >=*, >*
infixr 3  &&*
infixr 2  ||*
infixl 1 >>*, >>=*

-- The representation
newtype Exp a = E D.Exp deriving Eq

type IntE    = Exp Int
type DoubleE = Exp Double
type FloatE  = DoubleE
type BoolE   = Exp Bool

unE :: Exp a -> D.Exp
unE (E e) = e

mkE :: D.Exp -> Exp a
mkE e = E e

typ1 :: (D.Exp -> D.Exp) -> (Exp a -> Exp b)
typ2 :: (D.Exp -> D.Exp -> D.Exp) -> (Exp a -> Exp b -> Exp c)
typ3 :: (D.Exp -> D.Exp -> D.Exp -> D.Exp) -> (Exp a -> Exp b -> Exp c -> Exp d)
typ4 :: (D.Exp -> D.Exp -> D.Exp -> D.Exp -> D.Exp) -> (Exp a -> Exp b -> Exp c -> Exp d -> Exp e)
-- ...

typ1 f (E e1)                      = E (f e1)
typ2 f (E e1) (E e2)               = E (f e1 e2)
typ3 f (E e1) (E e2) (E e3)        = E (f e1 e2 e3)
typ4 f (E e1) (E e2) (E e3) (E e4) = E (f e1 e2 e3 e4)
-- ...


var :: ExpType a => Id -> Exp a
var n = e
 where
   e = E (D.Var n (expType e))

occFreeIn :: Id -> Exp a -> Bool
occFreeIn id (E e) = D.occFreeIn id e

class ExpType a where
  expType :: Exp a -> Type

instance ExpType Int        where expType _ = Int
instance ExpType ()         where expType _ = Void
instance ExpType Double     where expType _ = Double
instance ExpType Bool       where expType _ = Bool
instance ExpType V4.F32vec4 where expType _ = F32vec4

instance ExpType ty => ExpType (Action ty) where
	expType x = Action (expType (f x))
	  where
	    f :: Exp (Action a) -> Exp a
	    f = undefined

instance ExpType ty => ExpType (Arr ty) where
  expType arr = Arr (expType (arrBogusVal arr))
                 where
                   arrBogusVal :: ArrE a -> Exp a
                   arrBogusVal _ = error "arrBogusVal: evaluated"

instance ExpType ty => ExpType (Ref ty) where
  expType ref = Ref (expType (refBogusVal ref))
                 where
                   refBogusVal :: RefE a -> Exp a
                   refBogusVal _ = error "refBogusVal: evaluated"

instance (ExpType a, ExpType b) => ExpType (a -> b) where
  expType fun = Fun (expType (funArgTy fun)) (expType (funResTy fun))
  	         where
		  funArgTy :: Exp (a->b) -> Exp a
		  funArgTy = undefined

		  funResTy :: Exp (a->b) -> Exp b
		  funResTy = undefined


-- How much of the "Syntactic" machinery do I have to repeat?

class Syntactic a where
  -- Type-specific optimizing form
  ifE :: BoolE -> a -> a -> a

instance Syntactic (Exp a) where
  ifE              = typ3 SD.ifE

instance TCheck (Exp a) where
  tcheck str (E e) = E (tcheck str e)


instance Syntactic b => Syntactic (a -> b) where
  ifE c fa fb  = \ x -> ifE c (fa x) (fb x)

-- Syntactic tuples.
-- Just distribute the if.  Later the resulting Ifs will get extracted
-- upward and merged.

instance (Syntactic a, Syntactic b) => Syntactic (a,b) where
  ifE co (a,b) (a',b') = (ifE co a a', ifE co b b')

instance (Syntactic a, Syntactic b, Syntactic c) => Syntactic (a,b,c) where
  ifE co (a,b,c) (a',b',c') = (ifE co a a', ifE co b b', ifE co c c')

instance (Syntactic a, Syntactic b, Syntactic c, Syntactic d)
    => Syntactic (a,b,c,d) where
  ifE co (a,b,c,d) (a',b',c',d') = (ifE co a a', ifE co b b'
                                   ,ifE co c c', ifE co d d')


---- Booleans

(==*), (/=*) :: Eq a => Exp a -> Exp a -> BoolE

(==*) = typ2 (SD.==*)
(/=*) = typ2 (SD./=*)

(<*), (<=*), (>=*), (>*) :: Ord a => Exp a -> Exp a -> BoolE
(<*)  = typ2 (SD.<*)
(<=*) = typ2 (SD.<=*)
(>=*) = typ2 (SD.>=*)
(>*)  = typ2 (SD.>*)

(&&*), (||*) :: BoolE -> BoolE -> BoolE
(&&*) = typ2 (SD.&&*)
(||*) = typ2 (SD.||*)

notE :: BoolE -> BoolE
notE = typ1 SD.notE

b2i :: BoolE -> IntE
b2i = typ1 SD.b2i
-- The following def causes too much if floating
-- b2i b = ifE b 1 0

true  = E SD.true
false = E SD.false


-----  Numbers

-- Now we define a bunch of methods.  The naming policy is to use the
-- same name as the unlifted function if overloaded and the types permit.
-- Sadly, type restrictions prevent us from defining some of the methods
-- on behaviors.  In such cases, add a suffix of "*" or "E" if the name is
-- a syntactic "operator" rather than an identifier.

noOverloadE   :: String -> String -> a
noOverloadEId :: String -> a
noOverloadEOp :: String -> a

noOverloadE suffix name =
 error ("No overloading of \"" ++ name ++ "\" for expressions.  Use \"" ++
        name ++ suffix ++"\" instead.")

noOverloadEId = noOverloadE "E"
noOverloadEOp = noOverloadE "*"

instance  Ord (Exp Int)  where
  (<)  = noOverloadEOp "<"
  (<=) = noOverloadEOp "<="
  (>)  = noOverloadEOp ">"
  (>=) = noOverloadEOp ">="
  min  = typ2 min
  max  = typ2 max

instance Enum (Exp Int) where
  toEnum n = E (D.LitInt n)
  fromEnum (E (D.LitInt n)) = n
--   toEnum       = error "toEnum: no def for Exp"
--   fromEnum e   = error $ "fromEnum: no def for Exp " ++ show e
--   enumFrom     = error "enumFrom: no def for Exp"
--   enumFromThen = error "enumFromThen: no def for Exp"

instance Real (Exp Int) where
  toRational = error "toRational: not defined on Exp"

instance Num (Exp Int) where
  (+)         = typ2 (+)
  (*)         = typ2 (*)
  negate      = typ1 negate
  abs         = typ1 abs
  fromInteger = E . D.LitInt . fromInteger

instance Num (Exp Double) where
  (+)         = typ2 (+)
  (*)         = typ2 (*)
  negate      = typ1 negate
  abs         = typ1 abs
  fromInteger = E . D.LitDouble . fromInteger

instance Num (Exp V4.F32vec4) where
  (+) = typ2 (+)
  (*) = typ2 (*)
  negate = typ1 negate
  -- Add an abs?
--  fromInteger i = E . D.LitF32vec4 (V4.doubleToF32vec4 (fromInteger i))

instance Integral (Exp Int) where
  mod = typ2 mod
  div = typ2 div

  quot      = error "quot: no def for Exp"
  rem       = error "rem: no def for Exp"
  quotRem   = error "quotRem: no def for Exp"
  divMod    = error "divMod: no def for Exp"
  toInteger = error "toInteger: no def for Exp"
  toInt     = error "toInt: no def for Exp"

evenE, oddE :: IntE -> BoolE
evenE e = e `mod` 2 ==* 0
oddE    = notE . evenE


instance Fractional (Exp Double) where
  recip        = typ1 recip
--  fromDouble   = E . D.LitDouble . fromDouble
  fromRational = E . D.LitDouble . fromRational

instance Fractional (Exp V4.F32vec4) where
  recip        = typ1 recip
  -- fromDouble   = E . D.LitF32vec4 . fromDouble
  -- fromRational = E . D.LitF32vec4 . fromRational
  fromRational r = packF32 d d d d
    where
      d = fromRational r


instance Floating (Exp Double) where
  pi      = E pi
  sin     = typ1 sin
  cos     = typ1 cos
  logBase = typ2 logBase
  (**)    = typ2 (**)
  -- Add more later
  -- sqrt uses default.
  log = typ1 log
  exp = typ1 exp


instance  Ord (Exp Double)  where
  (<)  = noOverloadEOp "<"
  (<=) = noOverloadEOp "<="
  (>)  = noOverloadEOp ">"
  (>=) = noOverloadEOp ">="
  min  = typ2 min
  max  = typ2 max


instance Real (Exp Double)

instance RealFrac (Exp Double) where
    -- properFraction = noOverloadEId "properFraction"
    truncate       = noOverloadEId "truncate"
    round          = noOverloadEId "round"
    ceiling        = noOverloadEId "ceiling"
    floor          = noOverloadEId "floor"

truncateE,roundE, ceilingE, floorE :: DoubleE -> IntE

-- Note: If 'f2i' rounds towards minus infinity, use
-- the following set of defns for the rounding functions:
truncateE x = f2i x + b2i (x <* 0)
roundE    x = f2i (x + 0.5)
ceilingE  x = f2i x + 1 -- curious.
floorE    x = f2i x

{-
-- Note: If 'f2i' rounds towards zero, use
-- the following set of defns for the rounding functions:

truncateE x = f2i x
roundE    x = floorE (x + 0.5)
ceilingE  x = f2i x + b2i (x >* 0)
floorE    x = f2i x - b2i (x <* 0)

-- Note: ceilingE (resp floorE) gives the wrong value for positive (resp
-- negative) whole numbers.  For instance, floorE (-3) = -4 and
-- ceiling 3 = 4.  
-}



instance RealFloat (Exp Double) where
  atan2 = typ2 atan2

---- Int/float conversions

i2f :: IntE -> DoubleE
i2f = typ1 SD.i2f

f2i :: DoubleE -> IntE
f2i = typ1 SD.f2i

fracPart :: DoubleE -> DoubleE
fracPart x = x - i2f (floorE x)

---- Bit-level operations

instance Bits (IntE) where
  (.&.) = typ2 (.&.)
  (.|.) = typ2 (.|.)
  complement = typ1 complement

-- Left and right shift
(<<<), (>>>) :: IntE -> IntE -> IntE

(<<<) = typ2 (SD.<<<)
(>>>) = typ2 (SD.>>>)


---- F32vec4

type F32vec4E = Exp V4.F32vec4

-- We need an Eq instance so that F32vec4E an Eq type.
{-
instance Eq V4.F32vec4 where
 (==) = error "==: no def for F32vec4"
-}

-- N.B. Lower order floats come first.
packF32 :: FloatE -> FloatE -> FloatE -> FloatE -> F32vec4E
packF32 = typ4 SD.packF32

nthF32 :: IntE -> F32vec4E -> FloatE
nthF32 = typ2 SD.nthF32

sumF32 :: F32vec4E -> FloatE
sumF32 = typ1 SD.sumF32


---- Tuples

-- Can't do a statically typed mkTuple, so do a few specialized versions

pair :: Exp a -> Exp b -> Exp (a,b)
pair = typ2 (\ a b -> SD.mkTuple [a,b])

fstE :: Exp (a,b) -> Exp a
fstE = typ1 (SD.tupleElt 0)

sndE :: Exp (a,b) -> Exp b
sndE = typ1 (SD.tupleElt 1)



---- Misc

deref32 :: IntE -> IntE
deref32 = typ1 SD.deref32

deref8 :: IntE -> IntE
deref8 = typ1 SD.deref8


---- Low-level one-dimensional arrays.

-- For now, the Haskell Arr type is just a place-holder type.  Consider
-- replacing with a real Haskell array, so we can do some compile-time
-- array evaluation.

data Arr a  = MkArr  
type ArrE a = Exp (Arr a)

type ArrReader a = IntE -> Exp (Action a)
type ArrWriter a = IntE -> a -> ActionE

{-
class ArrayLike arr elem where
  readArr  :: arr -> IntE -> elem
  writeArr :: arr -> IntE -> elem -> ActionE

instance ArrayLike (ArrE a) (Exp a) where
  readArr (E arr) (E i) = E (D.ReadArr arr i)
  writeArr (E arr) (E i) (E val) = E (D.WriteArr arr i val)
-}

-- Low level read&write primitives.
readArr :: ArrE a -> IntE -> Exp a
readArr (E arr) (E i) = E (D.ReadArr arr i)

readMutArr :: ExpType a => ArrE a -> ArrReader a
readMutArr (E arr) (E i) = E (D.ReadMutArr arr i)

writeArr :: ArrE a -> ArrWriter (Exp a)
writeArr (E arr) (E i) (E val) = E (D.WriteArr arr i val)

{-
 We provide two ways to construct a new array - one that
 doesn't provide an initial element value and one that does.
-}
unsafeNewArr :: (ExpType a)
             => IntE
             -> Exp (Action (Arr a))
unsafeNewArr size = res
  where
   res = E (SD.newArr (unE size) ty)
   ty  = expType (tyExtract res)
   
   tyExtract :: Exp (Action (Arr a)) -> Exp a
   tyExtract = undefined

newArr :: (ExpType a)
       => IntE
       -> Exp a
       -> Exp (Action (Arr a))
newArr size val =
  unsafeNewArr size >>=* \ arr ->
  loop size (\ v -> writeArr arr v val) >>*
  returnE arr

-- Array comprehension.  Takes number of elements n and element function
-- f, and forms array [f 0, ..., f (n-1)].
mkArr :: ExpType a => IntE -> (IntE -> Exp a) -> ArrE a

{- -- Let's not use yet.  CSE has smarts about MkArr

mkArr size elem = 
 performAction (
   unsafeNewArr size >>=* \ arr ->
   loop size (\ v -> writeArr arr v (elem v)) >>*
   returnE arr)
-}

-- Old version:
mkArr (E size) elem = E (SD.mkArr size (unE . elem . E))
-- But we're phasing out the MkArr constructor

lambda :: (ExpType a, ExpType b) => (Exp a -> Exp b) -> Exp (a -> b)
lambda f = E (SD.lambda (expType exp_a) (unE . f . E))
 where
   -- Hack to synthesize a value of type Exp a.
   exp_a = error "exp_a" (f exp_a)

lambdaVar :: (ExpType a, ExpType b) => Id -> (Exp a -> Exp b) -> Exp (a -> b)
lambdaVar v f = E (SD.lambdaVar v (expType exp_a) (unE . f . E))
 where
   -- Hack to synthesize a value of type Exp a.
   exp_a = error "exp_a" (f exp_a)

rand :: Exp (Action Int)
rand = E D.Rand

data Ref  a = RefC
type RefE a = Exp (Ref a)

newRef :: (ExpType a) => Exp a -> Exp (Action (Ref a))
newRef init = E (SD.newRef (unE init))

readRef :: RefE a -> Exp (Action a)
readRef ref = E (SD.readRef (unE ref))

writeRef :: RefE a -> Exp a -> Exp (Action ())
writeRef ref newVal = E (SD.writeRef (unE ref) (unE newVal))

updateRef :: (ExpType a) => RefE a -> (Exp a -> Exp a) -> ActionE
updateRef ref updater = readRef ref >>=* \ v ->  writeRef ref (updater v)

---- Imperative stuff

type Action  a = D.Action a
type ActionE = Exp (Action ())

{-
  To execute an action, use 'performAction'. 
-}
performAction :: (ExpType a)
	      => Exp (Action a)
	      -> Exp a
performAction e = E (D.Perform (unE e))

{-
 Monadic bind, Exp style. Notice the restriction on the range of
 types that an action might return - it is only possible to return
 values that are supported at run-time. 
 
 One unfortunate consequence of this restriction is that we can't define
 an instance of Monad (=> we're not able to use the 'do' notation.)
-}
bindAction :: (ExpType a, ExpType b)
	   => Exp (Action a)
	   -> (Exp a -> Exp (Action b))
	   -> Exp (Action b)
bindAction m k = 
	E (D.Bind (unE m)
		  (D.Lambda x ty (unE (k (E (D.Var x ty)))))
          )
 where
  x  = unsafePerformIO (D.newId' "i")
  ty = getActionType (expType m)

noop :: ActionE
noop = E SD.noop

setInt :: IntE -> IntE -> ActionE
setInt = typ2 SD.setInt


returnE :: ExpType a => Exp a -> Exp (Action a)
returnE e = E (D.Return (unE e))

(>>=*) a b = bindAction a b

(>>*) :: (ExpType a, ExpType b)
      => Exp (Action a)
      -> Exp (Action b)
      -> Exp (Action b)
(>>*) a b = bindAction a (const b)

-- Given loop limit n and body f, do f 0; ...; f (n-1)
loop :: IntE -> (IntE -> ActionE) -> ActionE
loop (E limit) body = E (SD.loop limit (unE . body . E))

-- sum from low to high inclusive
loopLHInc :: IntE -> IntE -> (IntE -> ActionE) -> ActionE
loopLHInc low high f = loop (high-low+1) $ \ i -> f (low + i)

doProgress :: IntE -> IntE -> ActionE
doProgress = typ2 (SD.doProgress)

copy1DArray :: IntE -> IntE -> ActionE
copy1DArray = typ2 (SD.copy1DArray)

-- f 0 + ... + f (n-1)
sumLoop :: (ExpType a, Num (Exp a)) => IntE -> (IntE -> Exp a) -> Exp a
sumLoop n f =
 performAction $
  newRef 0 >>=* \ ref ->
    loop n (\ i -> updateRef ref (+ (f i))) >>*
  readRef ref

-- sum from low to high inclusive
sumLoopLHInc :: (ExpType a, Num (Exp a)) => IntE -> IntE -> (IntE -> Exp a) -> Exp a
sumLoopLHInc low high f =
  sumLoop (high-low+1) $ \ i -> f (low + i)

-- ToDo: generalize sumLoop, sumLoop to a fold


instance Show (Exp a) where
  showsPrec p (E e) = PPDExp.showsPrecDExp p e

ppExp :: Exp a -> PPDoc
ppExp (E e) = PPDExp.ppExp e


-----  Some other useful types

-- Byte fractions:

type ByteFrac  = Int
type ByteFracE = IntE

fromBF :: ByteFracE -> FloatE
fromBF bf = i2f bf / 255.0

toBF :: FloatE -> ByteFracE
toBF x = floorE (x * 255.0)

multBF :: ByteFracE -> ByteFracE -> ByteFracE
multBF a b = ((t >>> 8) + t) >>> 8
 where t = a * b + 128


-- A note about byte ordering and names.  Intel is little-endian.  It
-- confuses me no end, and I don't like that, but it's the way things are.
-- For this reason, we use names like "BGR24" to mean an integer whose
-- lowest-order byte is blue, second-lowest is green, third-lowest is red.
-- The fourth is zero. (conal)

-- Note that the fourth lowest byte is zero in the type "BGR24" but is
-- undefined in "BGR32".

type BGR32  = Int
type BGR32E = Exp BGR32

type BGRA32  = Int
type BGRA32E = Exp BGR32

type BGR24  = Int
type BGR24E = Exp BGR24

type RGB24  = Int
type RGB24E = Exp RGB24


-- Real number between zero and one inclusive.
type Fraction  = Double
type FractionE = Exp Fraction

-- Linear interpolation
-- lerp c a b = (1 - c) * a + c * b
lerp c a b = a + c * (b - a)

-- Bytes
type Byte  = Int
type ByteE = Exp Byte

-- Pack and unpack bytes.  Note little-endian bias.
pack :: ByteE -> IntE -> IntE
pack lo hi = (hi <<< 8) .|. lo

-- Extract low byte and high remaining int
unpack :: IntE -> (ByteE,IntE)
unpack ab = (ab .&. 255, ab >>> 8)

-- Unpack 32-bit int into four bytes, low first.  Note that (>>>) is
-- *signed* shift, so if abcd is "negative" (i.e., d>=128), we need to
-- mask even d.  Thus the final unpack.  The "x"s represent a trail of
-- zeros or ones, depending on whether d>=128.
unpack4 :: IntE -> (IntE,IntE,IntE,IntE)
unpack4 abcd = (a,b,c,d)
  where
    (a,bcdx) = unpack abcd
    (b,cdxx) = unpack bcdx
    (c,dxxx) = unpack cdxx
    (d,xxxx) = unpack dxxx

packL :: [ByteFracE] -> IntE
packL = foldr pack 0


(!!*) :: Syntactic a => [a] -> IntE -> a

[x] !!* _ = x
(x:xs) !!* n = ifE (n ==* 0) x (xs !!* (n-1))

andE, orE :: [BoolE] -> BoolE
andE      = foldr (&&*) true
orE       = foldr (||*) false

-- Given an angle in [-pi,pi] (e.g., from atan2), add 2 pi if negative.
remapAng :: FloatE -> FloatE
remapAng a  = a + 2*pi * i2f (b2i (a<*0))


---- Misc animation

type TimeE = FloatE


translateT, scaleT :: TimeE -> (TimeE -> a) -> (TimeE -> a)

-- Shift in time.  x>0 delays
translateT x anim = anim . (\ t -> t - x)

-- Stretch time.  x>1 stretches (slows), x<1 compresses (speeds)
scaleT x anim = anim . (\ t -> t / x)



-- Function liftings.  lift<n> turns an n-ary function into one that turns
-- n functions into a new function.  (Not just images.)  For constants,
-- just use the prelude's "const".
lift1 :: (a -> b) -> (p -> a) -> (p -> b)
lift2 :: (a -> b -> c) -> (p -> a) -> (p -> b) -> (p -> c)
lift3 :: (a -> b -> c -> d) -> (p -> a) -> (p -> b) -> (p -> c) -> (p -> d)

lift1 h f1 x       = h (f1 x)
lift2 h f1 f2 x    = h (f1 x) (f2 x)
lift3 h f1 f2 f3 x = h (f1 x) (f2 x) (f3 x)
