/*
 * 
 * The Pan compiler needs to call out from Haskell 
 * to perform a couple of tasks, such as:
 * reading in bitmaps, interfacing to the RTCG (LVM) [no longer the case],
 * packaging up of plugins. 
 *
 * This file contains the implementations of these functions;
 * the Haskell-callable stubs to them is generated from 
 * PanUtils.idl
 */

#ifdef _MSC_VER
#define COBJMACROS
#include <windows.h>
#include <OCIdl.h>
#include <olectl.h>
#include <shlobj.h>
#include <tchar.h>
#else

#include <windows.h>
#include <tchar.h>

#define STDAPICALLTYPE STDCALL
// If compiling with gcc, we don't have MIDL generated .h
// files at our disposal (e.g., for ocidl.h), so we'll have to
// improvise and type out the protos manually.
#include "PanUtilsSrcHelp.h"
#endif

#include <malloc.h>
#include <stdio.h>

// Sanity check.
#include "PanUtilsSrc.h"

#define HIMETRIC_INCH	2540

static int dir_unique = 0;

static void NewDirName(char* root, char* prefix, int i, char** psDirName);

//
//   FUNCTION: MkPluginDir(char*, char*)
//
//   PURPOSE: Create new temporary Pan plugin directory.
//
//   COMMENTS:
//
//        Given a root directory and a prefix, this function
//        will create a new directory in the root directory.
//        Directories are made unique by appending a four-digit
//        number. It returns the name of the new directory.
//
char* MkPluginDir(char* root, char* prefix)
{

  char* dirname;
  BOOL created_dir_ok = FALSE;
  DWORD dw = 0;

  dirname = (char*)malloc(sizeof(char)*512);

  if (dirname == NULL) {
    return NULL;
  }


  while (created_dir_ok == FALSE) {
    // generate new candidate directory name.
    NewDirName(root, prefix, dir_unique, &dirname);
    // try to use it.
    created_dir_ok = CreateDirectory(dirname, NULL);
    
    if (created_dir_ok == TRUE) {
      return dirname;
    } else if (created_dir_ok == FALSE && 
	       (dw = GetLastError()) == ERROR_ALREADY_EXISTS ) {
      dir_unique++;
    } else {
      // some other error; report error & bail out.
      FormatMessage(
		    FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
		    NULL,
		    dw,
		    MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
		    (char*)dirname,
		    512,
		    NULL);
      
      MessageBox(NULL, dirname, "MkPluginDir", MB_OK | MB_ICONERROR );
      return NULL;
    }
  }
  return dirname;
}

// Construct a plugin file path from parts; used by MkPluginDir().
static void
NewDirName(char* root, char* prefix, int i, char** psDirName)
{
    char* sep = root[strlen(root)-1] == '\\' ? "" : "\\";
    sprintf(*psDirName, "%s%s%s%04d", root, sep, prefix, i);
}


//
//   FUNCTION: ShellOpen(char*);
//
//   PURPOSE: Open file/folder
//
//   COMMENTS:
//
//        Given a file or directory name, launch it. In the case of
//        a folder, displays a shell explorer window containing it.
//
void
ShellOpen (char* lpFile)
{
  ShellExecute (NULL, "open", lpFile, NULL, NULL, SW_SHOWNORMAL);
  return ;
}

//
//   FUNCTION: LoadBMPHeader(char*, int*, int*, int*, int*);
//
//   PURPOSE: Load the header of a BMP
//
//   COMMENTS:
//
//        Given the name of a BMP file, tries to look at its header,
//        returning the attributes of interest to Pan.
//
BOOL
LoadBMPHeader
         ( /*[in]*/   char* lpszFile
	 , /*[out]*/  LONG* pWidth
	 , /*[out]*/  LONG* pHeight
	 , /*[out]*/  LONG* pPlanes
	 , /*[out]*/  LONG* pBitCount
	 , /*[out]*/  LONG* pCompression
	 , /*[out]*/  LONG* pClrUsed
	 , /*[out]*/  LONG* pClrImportant
	 )
{
  HANDLE h;
  BOOL res;
  BITMAPFILEHEADER bmf;
  DWORD bytesRead = 0;
  BITMAPINFOHEADER bih;

  h = CreateFile( lpszFile
  	        , GENERIC_READ 
		, FILE_SHARE_READ
		, NULL
		, OPEN_EXISTING
		, FILE_ATTRIBUTE_NORMAL
		, NULL
		);

  if ( INVALID_HANDLE_VALUE == h ) {
     return FALSE;
  }

  res = ReadFile(h, &bmf, sizeof(BITMAPFILEHEADER), &bytesRead, NULL);
  
  // Check whether the file header is reasonable before continuing:
  if ( FALSE == res || bytesRead != sizeof(BITMAPFILEHEADER) || 
       bmf.bfType != (WORD)((int)'B' + 256* (int)'M') ||
       bmf.bfReserved1 != 0 || bmf.bfReserved2 != 0 ) {
     CloseHandle(h);
     return FALSE;
  }

  // ReadFile() allegedly resets it for us, but don't rely on it.
  bytesRead = 0;
  
  res = ReadFile(h, &bih, sizeof(BITMAPINFOHEADER), &bytesRead, NULL);
  
  if (  FALSE == res ) {
     CloseHandle(h);
     return FALSE;
  }

  // Assign the [out] params.
  *pWidth  = bih.biWidth;
  *pHeight = bih.biHeight;
  *pPlanes = bih.biPlanes;
  *pBitCount = bih.biBitCount;
  *pCompression = bih.biCompression;
  *pClrUsed = bih.biClrUsed;
  *pClrImportant = bih.biClrImportant;

  // Close down and return back home.
  CloseHandle(h);
  return TRUE;
}

//
//   FUNCTION: LoadBMPBitsPal(char*, void*, void*);
//
//   PURPOSE: Load the palette and bits of a BMP.
//
//   COMMENTS:
//
//        Given the name of a BMP file, tries to look at its header,
//        returning the attributes of interest to Pan.
//
BOOL
LoadBMPBitsPal
         ( /*[in]*/   char* lpszFile
	 , /*[out]*/  RGBQUAD** pBits
	 , /*[out]*/  RGBQUAD** pPalette
	 )
{
  HANDLE h;
  BOOL res;
  BITMAPFILEHEADER bmf;
  DWORD bytesRead = 0;
  BITMAPINFOHEADER bih;
  DWORD size;
  DWORD offset;
  DWORD palSizeBytes;
  RGBQUAD* palette;

  h = CreateFile( lpszFile
  	        , GENERIC_READ 
		, FILE_SHARE_READ
		, NULL
		, OPEN_EXISTING
		, FILE_ATTRIBUTE_NORMAL
		, NULL
		);

  if ( INVALID_HANDLE_VALUE == h ) {
     return FALSE;
  }

  res = ReadFile(h, &bmf, sizeof(BITMAPFILEHEADER), &bytesRead, NULL);
  
  // Check whether the file header is reasonable before continuing:
  if ( FALSE == res || bytesRead != sizeof(BITMAPFILEHEADER) || 
       bmf.bfType != (WORD)((int)'B' + 256* (int)'M') ||
       bmf.bfReserved1 != 0 || bmf.bfReserved2 != 0 ) {
     CloseHandle(h);
     return FALSE;
  }
  
  size   = bmf.bfSize - sizeof(BITMAPFILEHEADER);
  offset = bmf.bfOffBits - sizeof(BITMAPFILEHEADER);
  
  res = ReadFile(h, &bih, sizeof(BITMAPINFOHEADER), &bytesRead, NULL);
  
  palSizeBytes = sizeof(RGBQUAD)*bih.biClrUsed;

  palette = (RGBQUAD*)malloc(palSizeBytes);
  
  if ( NULL == palette ) {
     CloseHandle(h);
     return FALSE;     
  }
  
  res = ReadFile(h, palette, palSizeBytes, &bytesRead, NULL);
  
  if ( FALSE == res || bytesRead != palSizeBytes ) {
	CloseHandle(h);
        free(palette);
	return FALSE;  
  }
  
  *pBits = (RGBQUAD*)malloc(size);
  
  if ( NULL == *pBits ) {
     CloseHandle(h);
     free(palette);
     return FALSE;     
  }

  res = ReadFile(h, *pBits, size, &bytesRead, NULL);

  if (  FALSE == res ) {
     CloseHandle(h);
     free(palette);
     free(*pBits);
     return FALSE;
  }
  
  *pPalette = (void*)palette;

  // Close down and return back home.
  CloseHandle(h);
  return TRUE;
}

int floatToInt(float x)
{
  union { float f; int i;} u;
  
  u.f = x;
  
  return u.i;
}

int doubleToInt(double x, int* y)
{
  typedef struct {int a; int b;} dW;
  union  {double d; dW i;} u;
  
  u.d = x;

  *y = u.i.b;

  return u.i.a;
}


/*
// Test whether the user's machine has SIMD FP.  Borrowed from the "Intel
// architecture optimization reference manual", page 3-3

BOOL hasSimdFP()
{
    int retVal = FALSE;
    _asm {
        push eax
        push ebx
        push ecx
        push edx
        mov eax, 1    ; request for feature flags
        cpuid         ; 0Fh, 0A2h cpuid instruction
        test EDX, 002000000h; bit 25 in feature flags equal to 1
        mov retVal, 0    ; assume no
        jz noSIMD
        mov retVal, 1
        noSIMD:
        pop edx
        pop ecx
        pop ebx
        pop eax
    }
    return retVal;
}
*/

//
// FUNCTION:  ResolveImageName
//
// PURPOSE:   If the name of an image import doesn't refer to
//            an existing file, try prepending the directory
//            path of a the 'standard' image import directory.
//
//            If that doesn't refer to an existing file either, return
//            (a copy) of the original name. 
//
// COMMENTS:  The caller is responsible for free()ing the returned string.
//
char*
ResolveImageName(const char* szFile)
{
  char* fullName;
   
  if ( strchr(szFile, '\\') == NULL &&
       strchr(szFile, '/')  == NULL ) {
    // only do this for filenames that aren't absolute or relative
    // paths.
    HKEY  panKey;
    static char* imageDirStr = NULL;
    static int memoised = 0;
    DWORD dirStrLen;
    LONG  lRes;
	
    if (!memoised) {
      // Open our (machine-wide) location in the registry:
      lRes = RegOpenKeyEx
	  ( HKEY_LOCAL_MACHINE
	  , TEXT("Software\\Pan")
	  , 0
	  , KEY_READ
	  , &panKey
          );
      if ( lRes == ERROR_SUCCESS ) {

	// To avoid looking up the registry value everytime 
	// LoadImportedImg() is called, we memoise it here.
	//
	// The downside of this is, of course, that dynamic changes
	// to the imageDir value won't get picked up. There's ways
	// to get around that (cf. RegNotifyChangeValue()), but doesn't
	// seem to be worth the extra bother at the moment.
	//
	// We also leak a little bit of memory
	if (imageDirStr == NULL) {
	  imageDirStr = (char*)malloc(sizeof(char) * 256);
	}
	dirStrLen = 255;

	// Look for "imageDir" string value
	lRes = 
	  RegQueryValueEx ( panKey
			  , TEXT("imageDir")
			  , NULL
			  , NULL
			  , (BYTE*)imageDirStr
			  , &dirStrLen
			  );
	if ( lRes == ERROR_SUCCESS ) {
	  memoised = 1;
	}

	RegCloseKey(panKey);

      } else {
	// Make a copy of the string
	fullName = (char*)malloc(sizeof(char)*(strlen(szFile) + 1));
	strcpy(fullName, szFile);
    
	return fullName;
      }
    }

    // Prepend the imageDir to the filename.
    fullName = (char*)malloc(sizeof(char) * (strlen(imageDirStr) + strlen(szFile) + 2));

    strcpy(fullName, imageDirStr);
    strcat(fullName, "\\");
    strcat(fullName, szFile);
	
    return fullName;

  } else {
    // The filename contained forward or backward slashes, so 
    // assume it is pointing to some legitimate file.
    
    // Make a copy of the string
    fullName = (char*)malloc(sizeof(char)*(strlen(szFile) + 1));
    strcpy(fullName, szFile);
    
    return fullName;
  }
}


// This function loads a file into an IStream.
int
GetPicDimensions(LPCTSTR szFile, int* pWidth, int* pHeight)
{
  HANDLE hFile;
  DWORD dwFileSize;
  LPVOID pvData = NULL;
  HDC hdc = GetDC(NULL);
  IPicture* pIPic;
  HGLOBAL hGlobal;
  DWORD dwBytesRead;
  BOOL bRead;
  LPSTREAM pstm;
  HRESULT hr;
  char*   szFileReal;

  // find the real filename, i.e., if the image import refers
  // to a file in the std. import directory, prepend the directory path.
  szFileReal = ResolveImageName(szFile);

  // open file
  hFile = CreateFile(szFileReal, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);

  free(szFileReal);

  if ( hFile == INVALID_HANDLE_VALUE) {
    return FALSE;
  }
  // get file size
  dwFileSize = GetFileSize(hFile, NULL);
  if ( dwFileSize == -1) {
    return FALSE;
  }

  // alloc memory based on file size
  hGlobal = GlobalAlloc(GMEM_MOVEABLE, dwFileSize);
  
  if ( hGlobal == NULL ) {
    return FALSE;
  }

  pvData = GlobalLock(hGlobal);
  if ( pvData == NULL ) {
    return FALSE;
  }

  dwBytesRead = 0;
  // read file and store in global memory
  bRead = ReadFile(hFile, pvData, dwFileSize, &dwBytesRead, NULL);

  if ( bRead == FALSE ) {
    return FALSE;
  }
  GlobalUnlock(hGlobal);
  CloseHandle(hFile);

  pstm = NULL;
  // create IStream* from global memory
  hr = CreateStreamOnHGlobal(hGlobal, TRUE, &pstm);
  
  if ( !SUCCEEDED(hr) || !pstm ) {
    return FALSE;
  }

  // Create IPicture from image file
  hr = OleLoadPicture(pstm, dwFileSize, FALSE, &IID_IPicture, (LPVOID *)&pIPic);
  IUnknown_Release((IUnknown*)pstm);
  if ( !SUCCEEDED(hr) ) {
    return FALSE;
  } else {
     // get width and height of picture
    long hmWidth;
    long hmHeight;
    IPicture_get_Width(pIPic,&hmWidth);
    IPicture_get_Height(pIPic,&hmHeight);
    // convert himetric to pixels
    *pWidth  = MulDiv(hmWidth, GetDeviceCaps(hdc, LOGPIXELSX), HIMETRIC_INCH);
    *pHeight = MulDiv(hmHeight, GetDeviceCaps(hdc, LOGPIXELSY), HIMETRIC_INCH);

    IUnknown_Release(pIPic);

    return TRUE;
  }
}

//
// It's awkward to set the initial directory for BrowseForFolder(),
// have to resort to using BrowseCallbackProc().  (from Q179378.)
// 
#ifdef _MSC_VER
static
int
CALLBACK
BrowseCallbackProc
     ( HWND hwnd
     , UINT uMsg
     , LPARAM lp
     , LPARAM lpData
     )
{
  switch(uMsg) {
  case BFFM_INITIALIZED: {
    // WParam is TRUE since you are passing a path.
    // It would be FALSE if you were passing a pidl.
    SendMessage(hwnd,BFFM_SETSELECTION,TRUE,(LPARAM)lpData);
  break;
  }
  case BFFM_SELCHANGED: {
      SendMessage(hwnd,BFFM_SETSTATUSTEXT,0,(LPARAM)lpData);
      break;
  }
  default:
    break;
  }
  return 0;
}
#endif

//
// FUNCTION: BrowseForFolder
//
// PURPOSE: Display the browse-for-folder dialog box, allowing
//          the user to pick a directory.
//
// 
void
BrowseForFolder ( char* title, char* initDir, char* pPath )
{
#ifndef _MSC_VER
  pPath[0] = '\0';
  return;
#else
    BROWSEINFO browseInfo;
    IMalloc* lpMalloc;
    LPITEMIDLIST lpItemIDList;
    
    pPath[0] = '\0';
    // needed by the task allocator
    OleInitialize(NULL);
    
    if (SHGetMalloc(&lpMalloc) != NOERROR )
       return; // failed to get allocator

    ZeroMemory(&browseInfo, sizeof(browseInfo));
    browseInfo.hwndOwner = NULL;  // start at the Desktop
    browseInfo.pidlRoot = NULL; 
    browseInfo.pszDisplayName = 0;
    browseInfo.lpszTitle = title;  // passed in
    browseInfo.ulFlags = BIF_DONTGOBELOWDOMAIN
#ifdef BIF_USENEWUI
	                   | BIF_USENEWUI
#endif
			   ;
    browseInfo.lpfn = BrowseCallbackProc;
    browseInfo.lParam = (LPARAM)initDir;
    
    if ((lpItemIDList = SHBrowseForFolder(&browseInfo)) != NULL) {
      SHGetPathFromIDList(lpItemIDList, pPath);
    }
    IMalloc_Free(lpMalloc, lpItemIDList);
    IMalloc_Release(lpMalloc);      
    return;
#endif
}

//
// FUNCTION: CreateNewGuid
//
// PURPOSE: Generate a new GUID, and return it in string form.
//
LPOLESTR
CreateNewGuid()
{
  GUID guid;
  HRESULT hr;
  LPOLESTR szGuid;
  
  hr = CoCreateGuid(&guid);
  
  if (FAILED(hr)) {
    MessageBox(NULL, "CreateNewGuid failed", "CreateNewGuid", MB_OK);
    return NULL;
  } else {
    szGuid = (LPOLESTR)malloc(sizeof(OLECHAR)*40); // should just about be big enough to hold a GUID string.
    if (szGuid) {
      StringFromGUID2(&guid, szGuid, 40);
    }
    return szGuid;
  }
}

//
// FUNCTION: StrToHive
//
// PURPOSE:  It's convenient to use shorthand for the
//           standard hives (i.e., "HKLM" is HKEY_LOCAL_MACHINE.)
//           This function implements the mapping.
//
static
HKEY
StrToHive ( const char* hive )
{
  if (strcmp(hive, "HKLM") == 0) {
    return HKEY_LOCAL_MACHINE;
  }
  if (strcmp(hive, "HKCU") == 0) {
    return HKEY_CURRENT_USER;
  }
  if (strcmp(hive, "HKCU") == 0) {
    return HKEY_CURRENT_USER;
  }
  if (strcmp(hive, "HKU") == 0) {
    return HKEY_USERS;
  }
  return (HKEY)NULL;
}

// Maximum length of a string read out of Reg. by GetRegistryStr().
#define MAX_REG_SZ_LEN  255

//
// FUNCTION: ReadRegistryStr
//
// PURPOSE: Convenient helper for reading a string value out of
//          the Registry.
//
// COMMENTS: Result string is allocated using malloc(). The max
//           length of the string is currently hard-wired.
//
char*
ReadRegistryStr ( /*[in]*/ const char* hive
	        , /*[in]*/ const char* path
	        , /*[in]*/ const char* key
	        )
{
  HKEY hHive = StrToHive(hive);
  HKEY hKey;
  LONG lRes;
  char* valStr;
  DWORD valStrLen;
  
  lRes = RegOpenKeyEx 
                 ( hHive
                 , path
		 , 0
		 , KEY_READ
		 , &hKey
		 );
  if ( lRes == NO_ERROR ) {
    valStr    = (char*)malloc(sizeof(char)* (MAX_REG_SZ_LEN + 1));
    valStrLen = MAX_REG_SZ_LEN;
  
    lRes = RegQueryValueEx
                 ( hKey
	   	 , key
		 , NULL
		 , NULL
		 , (BYTE*)valStr
		 , &valStrLen
		 );
		 
    RegCloseKey(hKey);
    RegCloseKey(hHive);

    if ( lRes != NO_ERROR ) {
      free(valStr);
      return NULL;
    } else {
      return valStr;
    } 
  } else {
    return NULL;
  }
}

//
// FUNCTION: WriteRegistryStr
//
// PURPOSE: Convenient helper for writing a string value to the registry.
//
HRESULT
WriteRegistryStr ( /*[in]*/ const char* hive
	         , /*[in]*/ const char* path
	         , /*[in]*/ const char* key
		 , /*[in]*/ const char* val
	         )
{
  HKEY hHive = StrToHive(hive);
  HKEY hKey;
  LONG lRes;
  DWORD valStrLen = strlen(val);
  
  lRes = RegOpenKeyEx 
                 ( hHive
                 , path
		 , 0
		 , KEY_ALL_ACCESS
		 , &hKey
		 );
  if ( lRes == NO_ERROR ) {

    lRes = RegSetValueEx
                 ( hKey
	   	 , key
		 , 0
		 , REG_SZ
		 , (BYTE*)val
		 , valStrLen
		 );

    RegCloseKey(hKey);
    RegCloseKey(hHive);

    if ( lRes != NO_ERROR ) {
      return HRESULT_FROM_WIN32(lRes);
    } else {
      return S_OK;
    }
  } else {
    return HRESULT_FROM_WIN32(lRes);
  }
}

//
// FUNCTION: AllocRegistryKey
//
// PURPOSE: If not already present, creates a registry key.
//
// COMMENTS: Friendly wrapper around RegCreateKeyEx().
//
HRESULT
AllocRegistryKey
            ( /*[in]*/ const char* hive
	    , /*[in]*/ const char* path
	    , /*[in]*/ const char* dVal
	    )
{
  HKEY hHive = StrToHive(hive);
  HKEY hKey;
  LONG lRes;
  DWORD keyDispo;
  
  lRes = RegCreateKeyEx 
                 ( hHive
                 , path
		 , 0
		 , NULL
		 , REG_OPTION_NON_VOLATILE
		 , KEY_ALL_ACCESS
		 , NULL
		 , &hKey
		 , &keyDispo
		 );

  if ( lRes == NO_ERROR ) {
    if ( dVal != NULL ) {
      // Set the default value of the key.
      lRes = RegSetValueEx 
	          ( hKey
		  , NULL
		  , 0
		  , REG_SZ
		  , (BYTE*)dVal
		  , strlen(dVal)
		  );
      RegCloseKey(hKey);
      return HRESULT_FROM_WIN32(lRes);
    } else {
      RegCloseKey(hKey);
      return S_OK;
    }
  } else {
    return HRESULT_FROM_WIN32(lRes);
  }
}

// number of chars needed to hold text rep. {xxxx-xx-xx-xx-xxxx}
// of a CLSID.
#define CLSID_STRLEN 38

typedef enum {
    InProcServer,
    InProcHandler,
    LocalServer
} ServerKind;

//
// FUNCTION:   CLSIDToBinaryPath(CLSID* rclsid, CTSTR* compKind)
//
// PURPOSE:    Given a CLSID, resolve it to the binary implementing
//             that component. 
//
// COMMENTS:   The responsibility of releasing the result string
//             is transferred to the caller. Release it using free().

LPTSTR
CLSIDToBinaryPath
    ( REFCLSID   rclsid
    , ServerKind sKind
    )
{
    HKEY        hKey;
    LONG        lRes;
    LPTSTR      subKeyStr;
    LPCTSTR     clsidPrefix = TEXT("CLSID");
    LPCTSTR     compKind;
    LPOLESTR    clsidString;
    int         subKeyStringLen;
    int         v;
    DWORD       dwType;
    DWORD       bufSize;
    LPTSTR      resString = NULL;
    HRESULT     hr;
    LPTSTR      clsidStringT;
    int         len;
    
    // construct the right registry key for the component

    switch (sKind) {
        // here's the reason why I miss having a CLSID -> binary component
        // resolver provided by the COM library - you have to emulate what
        // the SCM does (and, who knows, whether this will work with newer
        // versions of the SCM (e.g., for Win64)?)
    case InProcServer:  compKind = TEXT("InprocServer32"); break;
    case InProcHandler: compKind = TEXT("InprocHandler32"); break;
    case LocalServer:   compKind = TEXT("LocalServer32"); break;
    }

    subKeyStringLen = _tcslen(clsidPrefix) +
                      CLSID_STRLEN +
                      4 + _tcslen(compKind);

    subKeyStr = (LPTSTR)_alloca(sizeof(TCHAR) * subKeyStringLen);

    // convert the CLSID into "{xxxx-xx-xx-xx-xxxx}" form
    hr = StringFromCLSID(rclsid, &clsidString);

    if (FAILED(hr)) {
        return NULL;
    }

#ifdef UNICODE
    clsidStringT = clsidString;
#else
    len = WideCharToMultiByte (CP_ACP, 0, clsidString, -1, NULL, 0, NULL, NULL);
    clsidStringT = (char*)malloc(sizeof(char)*(len+1));
    
    WideCharToMultiByte ( CP_ACP, 0, clsidString, -1, clsidStringT, len, NULL, NULL);
#endif

    v = _sntprintf(subKeyStr, subKeyStringLen, TEXT("%s\\%s\\%s"), clsidPrefix, clsidStringT, compKind);

#ifndef UNICODE
    free(clsidStringT);
#endif

    CoTaskMemFree(clsidString);

    if (v < 0) { // not enough room in subKeyStr
        //ToDo: use ReportErrorMessage() utility function.
        return NULL;
    }

    // use the registry to locate the DLL that implements
    // the server.

    lRes = RegOpenKeyEx(HKEY_CLASSES_ROOT, subKeyStr, 0, KEY_READ, &hKey);

    if (lRes != ERROR_SUCCESS) {
        return NULL;
    }


    // figure out the size of the key's unnamed value
    bufSize = 0; dwType = REG_SZ;
    lRes = RegQueryValueEx(hKey, NULL, NULL, &dwType, NULL, &bufSize);

    if (lRes == ERROR_SUCCESS) {
        // Q: does the 'bufSize' include the term. zero?
        resString = (LPTSTR)malloc(bufSize+1);

        lRes = RegQueryValueEx (hKey, NULL, NULL, &dwType, (LPBYTE)resString, &bufSize);

        if ( lRes != ERROR_SUCCESS || dwType != REG_SZ) {
            free(resString);
            resString = NULL;
        }
    }

    // close up and return
    RegCloseKey(hKey);
    return resString;

}


//
// FUNCTION:    HRESULT UnregisterServer(CLSID* rclsid)
// 
// PURPOSE:     Given a CLSID, unregister the inproc server
//              associated with (if any) on the local machine.
//
// COMMENTS:    
HRESULT
UnregisterServer
    ( CLSID* rclsid )
{
    HRESULT hr = S_OK;

    // locate the path to the server.
    LPTSTR compPath = CLSIDToBinaryPath(rclsid, InProcServer);

    // mimic regsvr32 and call DllUnregisterServer() on the DLL.
    if (compPath != NULL) {
        HINSTANCE hDll = LoadLibrary(compPath);

        if (hDll != NULL) {
            HRESULT (STDAPICALLTYPE *pfn)();
            pfn = GetProcAddress(hDll, "DllUnregisterServer");

            if (pfn != NULL) {
                hr = pfn();
            }
            FreeLibrary(hDll);
        }
    }
    
    free(compPath);
    return hr;

}


//
// FUNCTION:  UnregisterProgID
//
// PURPOSE:   Given a (version-indep) ProgID, check to see
//            whether it already has got an installed component
//            associated with it. If so, unregister the server/component.
//
void
UnregisterProgID(char* progID)
{
  HRESULT hr;
  WCHAR*  wProgID;
  int     wLen;
  CLSID   clsid;

  // Unfortunately, since we want the code to compile with VC++
  // and gcc (via cygwin), we have to do char*->OLESTR conversion
  // via the least common denominator. Tiresome.
  //
  wLen = MultiByteToWideChar(CP_ACP, 0, progID, -1, NULL, 0);
  if ( wLen < 0 ) {
    return;
  }

  wProgID = (WCHAR*)malloc(sizeof(WCHAR)*(wLen+1));

  MultiByteToWideChar(CP_ACP, 0, progID, -1, wProgID, wLen); 
  wProgID[wLen]=0;

  hr = CLSIDFromProgID(wProgID, &clsid);

  if (SUCCEEDED(hr)) {
    UnregisterServer(&clsid);
  }
  free(wProgID);
}

//
// FUNCTION:  char* getShortPathName(char*)
//
// PURPOSE:   Given a file path, convert into 8.3 form, i.e., a
//            wrapper around GetShortPathName().
//
char*
getShortPathName(const char* sPath)
{
  DWORD res;
  char* sShort = (char*)malloc(sizeof(char)*MAX_PATH);

  if (sShort) {
    if ((res = GetShortPathNameA(sPath,sShort,MAX_PATH * sizeof(char))) == 0 || res > MAX_PATH ) {
      // In the case of (v. unlikely) failure, make sure the
      // result string is the empty string.
      sShort[0] = '\0';
    }
  }
  return sShort;
}

//
// FUNCTION:  char* getTempPath()
//
// PURPOSE:   Return the path of local temp. directory, i.e., a
//            wrapper around GetShortPathName().
//
char*
getTempPath(const char* sPath)
{
  DWORD res;
  
  char* tempPath = (char*)malloc(sizeof(char)*MAX_PATH);

  if (tempPath) {
    if ((res = GetTempPath(MAX_PATH, tempPath) == 0)) {
      // In the case of (v. unlikely) failure, make sure the
      // result string is the empty string.
      tempPath[0] = '\0';
    }
  }
  return tempPath;
}

